from itpminer.utils.components import ErrorChecker
from itpminer.utils.types import *


def create(time_rel: int, item) -> ExtendedItemType:
    ErrorChecker.check_type(time_rel, "time_rel", int)
    return (time_rel, item)


def check_type(extended_item: ExtendedItemType):
    ErrorChecker.check_type(extended_item, "extended_item", tuple)
    ErrorChecker.check_type(extended_item[0], "extended_time[0]", int)


def display(extended_item: ExtendedItemType) -> str:
    check_type(extended_item)
    return f"{extended_item[1]}({extended_item[0]})"



from itpminer.utils.types import *
import networkx as nx
import matplotlib.pyplot as plt


def rules_graph(rules_display_dict: RulesDisplayDictType, rules_dict: RulesDictType, weighted: bool = False, one_to_one: bool = False):
    rules_one_to_one_dict: RulesDictType = {}
    if one_to_one:
        for key, item in rules_dict.items():
            if len(item.A) == 1 and item.A[0][0] == item.B[0][0] == 0:
                rules_one_to_one_dict[key] = item

        rules_one_to_one_display_dict: RulesDisplayDictType = {}
        for key, item in rules_one_to_one_dict.items():
            rules_one_to_one_display_dict[key] = RuleDisplay(
                A=f"{item.A[0][1]}",
                sup_A=item.sup_A,
                B=f"{item.B[0][1]}",
                sup_B=item.sup_B,
                A_and_B=f"{item.A}&{item.B}",
                sup_A_and_B=item.sup_A_and_B,
                conf=item.conf,
                lift=item.lift
            )
        rules_display_dict = rules_one_to_one_display_dict

    plt.figure(figsize=(20, 10))
    G = nx.DiGraph()
    edge_labels = {}

    for rule in rules_display_dict.values():
        A_node = f"{rule.A}\nS:{round(rule.sup_A, 2)}"
        B_node = f"{rule.B}\nS:{round(rule.sup_B, 2)}"
        edge = (A_node, B_node)
        G.add_edge(*edge, weight=rule.lift)
        edge_labels[edge] = f"L: {round(rule.lift, 2)}, C: {round(rule.conf, 2)}"

    pos = nx.circular_layout(G)
    nx.draw(G, pos, node_size=5000, node_color='pink', alpha=0.9)
    nx.draw_networkx_labels(G, pos)
    nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels)
    if weighted:
        for edge in G.edges(data='weight'):
            nx.draw_networkx_edges(
                # There's a false positive test from mypy
                G, pos, edgelist=[edge], width=edge[2])  # type: ignore
    plt.show()


from itpminer.utils.types import *
from itpminer.utils.components import Pattern
import pandas as pd


def association_rules(tree_dict: TreeDictType, min_conf=0.1, min_lift=1) -> Tuple[RulesDictType, RulesDisplayDictType, pd.DataFrame]:
    root_node = tree_dict[()]
    database_len = len(root_node.value[1])
    rules_dict: RulesDictType = {}
    rules_display_dict: RulesDisplayDictType = {}

    def support(node):
        return len(node.value[1]) / database_len

    def antecedent_traverse(node_antecedent: "Node" = root_node):
        for node_consequent in node_antecedent.children.values():
            if node_antecedent is not root_node:
                consequent_last_extended_item = node_consequent.value[0][-1]
                pattern_antecedent = node_antecedent.value[0]
                pattern_consequent = ((0, consequent_last_extended_item[1]),)

                support_antecedent_and_consequent = support(node_consequent)
                support_antecedent = support(node_antecedent)
                support_consequent = support(tree_dict[pattern_consequent])

                confidence = support_antecedent_and_consequent / support_antecedent
                if confidence >= min_conf:
                    lift = confidence / support_consequent
                    if lift >= min_lift:
                        rules_dict[node_consequent.value[0]] = Rule(
                            A=pattern_antecedent,
                            sup_A=support_antecedent,
                            B=(node_consequent.value[0][-1],),
                            sup_B=support_consequent,
                            A_and_B=node_consequent.value[0],
                            sup_A_and_B=support_antecedent_and_consequent,
                            conf=confidence,
                            lift=lift
                        )
                        rules_display_dict[node_consequent.value[0]] = RuleDisplay(
                            A=Pattern.display(pattern_antecedent),
                            sup_A=support_antecedent,
                            B=Pattern.display((node_consequent.value[0][-1],)),
                            sup_B=support_consequent,
                            A_and_B=Pattern.display(node_consequent.value[0]),
                            sup_A_and_B=support_antecedent_and_consequent,
                            conf=confidence,
                            lift=lift
                        )
            antecedent_traverse(node_consequent)

    antecedent_traverse()

    return rules_dict, rules_display_dict, pd.DataFrame(rules_display_dict.values())



from typing import Iterable, Optional, Dict
from itpminer.utils.components import ErrorChecker, DatList
from itpminer.utils.types import DatListType, PatternType


class Node():
    def __init__(self, dat_list: DatListType, parent: Optional["Node"] = None):
        if parent is not None:
            self.check_type(parent)
        self.parent = parent

        DatList.check_type(dat_list)
        self.value = dat_list

        self.children: Dict[PatternType, "Node"] = {}

    def check_type(self, node: Optional["Node"]):
        if node is not None:
            ErrorChecker.check_type(node, "Node", Node)

    def add_child(self, child: "Node"):
        self.check_type(child)
        self.children[child.value[0]] = child

    def add_children(self, children: Iterable["Node"]):
        isinstance(children, tuple)
        for child in children:
            self.check_type(child)
            self.add_child(child)

    def __repr__(self) -> str:
        return DatList.display(self.value) if self.value else str(None)



# Imports
from collections import defaultdict
import pandas as pd
from typing import Dict, List
from itpminer.utils.components import Pattern
from itpminer.utils.components.Node import Node
from itpminer.utils.types import *


def join_2(alpha: PatternType, beta: PatternType, tree_dict: TreeDictType, H2: H2Type, min_freq: int, max_span: int):
    tree_dict = tree_dict.copy()
    for pattern in [alpha, beta]:
        Pattern.check_type(pattern)
        if len(pattern) != 1:
            raise Exception("Pattern is not of length 1")

    # candidate_datlist_dict has (key: pattern, value: dats)
    candidate_datlist_dict = defaultdict(set)

    alpha_extended_item = alpha[0]
    beta_extended_item = beta[0]
    for w in range(0, max_span):
        pattern_2 = (alpha_extended_item, (w, beta_extended_item[1]))
        if pattern_2 in H2:
            if H2[pattern_2] >= min_freq:
                alpha_dats = tree_dict[alpha].value[1]
                beta_dats = tree_dict[beta].value[1]

                for alpha_dat in alpha_dats:
                    if alpha_dat + w in beta_dats:
                        candidate_datlist_dict[pattern_2].add(alpha_dat)

    for candidate_pattern, candidate_dats_set in candidate_datlist_dict.items():
        if len(candidate_dats_set) >= min_freq:
            node = Node(parent=tree_dict[alpha], dat_list=(
                candidate_pattern, tuple(candidate_dats_set)))
            tree_dict[alpha].add_child(node)
            tree_dict[candidate_pattern] = node
    return tree_dict


def join_k(alpha: PatternType, beta: PatternType, tree_dict: TreeDictType, H2: H2Type, min_freq: int):
    tree_dict = tree_dict.copy()
    for pattern in [alpha, beta]:
        Pattern.check_type(pattern)

    time_alpha_last, item_alpha_last = alpha[-1]
    time_beta_last, item_beta_last = beta[-1]
    pair_of_last = ((0, item_alpha_last),
                    (time_beta_last-time_alpha_last, item_beta_last))
    if H2[pair_of_last] >= min_freq:
        theta_pattern = Pattern.join(alpha, beta)

        alpha_times = tree_dict[alpha].value[1]
        beta_times = tree_dict[beta].value[1]
        theta_times = tuple(set(alpha_times).intersection(set(beta_times)))
        if len(theta_times) >= min_freq:
            theta_dat_list = (theta_pattern, theta_times)
            node = Node(parent=tree_dict[alpha], dat_list=theta_dat_list)
            tree_dict[alpha].add_child(node)
            tree_dict[theta_pattern] = node

    return tree_dict


def DFS(alpha: PatternType, frequent_patterns: FrequentPatternsType, tree_dict: TreeDictType, H2, min_freq: int, max_span: int):
    tree_dict = tree_dict.copy()
    frequent_patterns.append(alpha)

    parent = tree_dict[alpha[:-1]]

    siblings_dict = parent.children
    siblings_patterns = sorted(list(siblings_dict.keys()))
    alpha_index = siblings_patterns.index(alpha)

    J_alpha = siblings_patterns[alpha_index+1:]
    for pattern in J_alpha:
        tree_dict = join_k(alpha, pattern, tree_dict, H2, min_freq)

    E_alpha = sorted(list(tree_dict[alpha].children.keys()))
    for pattern in E_alpha:
        tree_dict = DFS(pattern, frequent_patterns,
                        tree_dict, H2, min_freq, max_span)

    return tree_dict


FrequentPatternsListType = List[Tuple[PatternType, int]]
FrequentPatternsDictType = Dict[PatternType, int]


def itp_miner(database: List[List], min_sup: float = 0.5, max_span: int = 2) -> Tuple[TreeDictType, FrequentPatternsDictType, FrequentPatternsListType, pd.DataFrame]:
    database = [sorted(list(set(itemset))) for itemset in database]
    min_freq: int = int(min_sup * len(database))

    # Key: pattern, Value: list of dats
    candidate_patterns = defaultdict(set)

    frequent_1_patterns: FrequentPatternsType = []
    root_node = Node(dat_list=((), tuple(range(len(database)))))
    tree_dict: TreeDictType = {(): root_node}
    H2: H2Type = defaultdict(int)

    for t, itemset in enumerate(database):
        for ind, item in enumerate(itemset):
            candidate_patterns[item].add(t)

            for item_next in database[t][ind+1:]:
                H2[((0, item), (0, item_next))] += 1

            for t_next in range(t+1, min(len(database), t+max_span)):
                for item_next in database[t_next]:
                    H2[((0, item), (t_next-t, item_next))] += 1

    for item, dats in candidate_patterns.items():
        if len(dats) >= min_freq:
            node = Node(parent=root_node, dat_list=(((0, item),), tuple(dats)))
            root_node.add_child(node)
            tree_dict[((0, item),)] = node
            frequent_1_patterns.append(((0, item),))

    frequent_patterns = frequent_1_patterns.copy()

    for alpha in frequent_1_patterns:
        J_alpha = frequent_1_patterns
        for pattern in J_alpha:
            tree_dict = join_2(alpha, pattern, tree_dict,
                               H2, min_freq, max_span)

        E_alpha = sorted(list(tree_dict[alpha].children.keys()))
        for pattern in E_alpha:
            tree_dict = DFS(pattern, frequent_patterns,
                            tree_dict, H2, min_freq, max_span)

    frequent_patterns_dict: FrequentPatternsDictType = {(): len(database)}
    for pattern in tree_dict:
        if pattern != ():
            frequent_patterns_dict[pattern] = len(tree_dict[pattern].value[1])

    frequent_patterns_list: FrequentPatternsListType = list(
        frequent_patterns_dict.items())

    frequent_patterns_df = pd.DataFrame(
        frequent_patterns_list, columns=["Pattern", "Support"])

    return tree_dict, frequent_patterns_dict, frequent_patterns_list, frequent_patterns_df



from typing import Tuple
from itpminer.utils.components import ErrorChecker, Pattern
from itpminer.utils.types import *


def check_type(dat_list: DatListType):
    pattern, time_abs_tuple = dat_list
    Pattern.check_type(pattern)
    ErrorChecker.check_type(time_abs_tuple, "time_abs_tuple", tuple)
    for time_abs in time_abs_tuple:
        ErrorChecker.check_type(time_abs, "time_abs", int)


def create(pattern: PatternType, time_abs_tuple: Tuple[int, ...]) -> DatListType:
    check_type((pattern, time_abs_tuple))
    return (pattern, time_abs_tuple)


def display(dat_list: DatListType) -> str:
    check_type(dat_list)
    return f"{Pattern.display(dat_list[0])}<{', '.join(map(str, dat_list[1]))}>"



def check_type(input, input_str: str, input_type: type):
    if not isinstance(input, input_type):
        raise Exception(
            f"{input_str} needs to be of type {input_type} but {input} is of type: {type(input)}")



from typing import Tuple
from itpminer.utils.components import ErrorChecker, ExtendedItem
from itpminer.utils.types import *


def check_type(pattern: PatternType):
    ErrorChecker.check_type(pattern, "pattern", tuple)
    for extended_item in pattern:
        ExtendedItem.check_type(extended_item)


def check_subset(subset: PatternType, superset: PatternType) -> bool:
    for s in [subset, superset]:
        check_type(s)
    return set(subset).issubset(superset)


def check_joinable(pattern1: PatternType, pattern2: PatternType) -> bool:
    for s in [pattern1, pattern2]:
        check_type(s)
    if len(pattern1) == len(pattern2):
        if len(pattern1) == 1:
            return True
        elif pattern1[:-1] == pattern2[:-1] and pattern1[-1] < pattern2[-1]:
            return True
    return False


def join(pattern1: PatternType, pattern2: PatternType) -> PatternType:
    for s in [pattern1, pattern2]:
        check_type(s)
    if not check_joinable(pattern1, pattern2):
        raise Exception(f"{pattern1} and {pattern2} are not joinable!")
    else:
        return (*pattern1, pattern2[-1])


def display(pattern: PatternType) -> str:
    check_type(pattern)
    return f"{{{', '.join([ExtendedItem.display(extended_item) for extended_item in pattern])}}}"



from typing import Dict, List, TYPE_CHECKING
from dataclasses import dataclass

if TYPE_CHECKING:
    from itpminer.utils.components.Node import Node

from typing import Any, Tuple

ExtendedItemType = Tuple[int, Any]
PatternType = Tuple[ExtendedItemType, ...]
DatListType = Tuple[PatternType, Tuple[int, ...]]


TreeDictType = Dict[PatternType, "Node"]
H2Type = Dict[PatternType, int]
FrequentPatternsType = List[PatternType]


@dataclass
class Rule:
    A: PatternType
    sup_A: float
    B: PatternType
    sup_B: float
    A_and_B: PatternType
    sup_A_and_B: float
    conf: float
    lift: float


@dataclass
class RuleDisplay:
    A: str
    sup_A: float
    B: str
    sup_B: float
    A_and_B: str
    sup_A_and_B: float
    conf: float
    lift: float


RulesDictType = Dict[PatternType, Rule]
RulesDisplayDictType = Dict[PatternType, RuleDisplay]
